@@ -25,18 +25,17 @@
 #  Frédéric MARTIN frederic.martin.fma@gmail.com
 #
 
-from __future__ import unicode_literals
+
 
 import os
 import sys
 import logging
 import base64
 import copy
-from ordereddict import OrderedDict
-import ConfigParser
+from collections import OrderedDict
+import configparser
 import argparse
-import types
-import locale
+from argparse import ArgumentError
 
 # -----------------------------------------------------------------------------
 # global logger variable
@@ -47,7 +46,7 @@ import locale
 DEFAULT_LOGGING_FORMAT = logging.Formatter(
     "%(asctime)s %(levelname)-8s: %(message)s", "%H:%M:%S")
 DEBUG_LOGGING_FORMAT = logging.Formatter(
-    "%(asctime)s %(levelname)-8s %(name)s:%(funcName)s:%(message)s",
+    "%(asctime)s %(levelname)-8s %(module)s:%(name)s:%(funcName)s:%(lineno)d:%(message)s",
     "%H:%M:%S")
 # logger handlers
 # pylint: disable-msg=C0103
@@ -121,12 +120,12 @@ class SectionHook(object):
             raise TypeError("First argument should be a subclass of _Section.")
         self.section = section
 
-        if not isinstance(attribute, types.UnicodeType):
+        if not isinstance(attribute, str):
             raise TypeError("Second argument should be a string, "
                             + "attribute name.")
         self.attribute = attribute
 
-        if not isinstance(opt_name, types.UnicodeType):
+        if not isinstance(opt_name, str):
             raise TypeError("Third argument should be a string, option name.")
         self.opt_name = opt_name
 
@@ -155,7 +154,7 @@ class Config(object):
         self.sections = OrderedDict()
         self._default_section = self.add_section(SimpleSection("DEFAULT"))
         self.parser = None
-        self.file_parser = ConfigParser.SafeConfigParser()
+        self.file_parser = configparser.SafeConfigParser()
 
     def add_section(self, section):
         """Add a new Section object to the config. Should be a subclass of
@@ -165,6 +164,11 @@ class Config(object):
         self.sections[section.get_key_name()] = section
         return section
 
+    def get_section(self, name):
+        if name.lower() == "default":
+            return self._default_section
+        return self.sections.get(name)
+
     def get_default_section(self):
         """This method will return default section object"""
         return self._default_section
@@ -182,7 +186,7 @@ class Config(object):
         log = logging.getLogger('argtoolbox')
         discoveredFileList = []
         if self.config_file:
-            if isinstance(self.config_file, types.UnicodeType):
+            if isinstance(self.config_file, str):
                 discoveredFileList = self.file_parser.read(self.config_file)
             else:
                 discoveredFileList = self.file_parser.readfp(
@@ -207,14 +211,14 @@ class Config(object):
 
         log.debug("loading configuration ...")
         if exit_on_failure:
-            for s in self.sections.values():
+            for s in list(self.sections.values()):
                 log.debug("loading section : " + s.get_section_name())
                 try:
                     s.load(self.file_parser)
                 except ValueError:
                     sys.exit(1)
         else:
-            for s in self.sections.values():
+            for s in list(self.sections.values()):
                 log.debug("loading section : " + s.get_section_name())
                 s.load(self.file_parser)
 
@@ -245,10 +249,12 @@ class Config(object):
         4. reloading configuration using cli argument like a configuration
         file name.
         """
+        #from argcomplete import debug
         # Parsing the command line looking for the previous options like
         # configuration file name or server section. Extra arguments
         # will be store into argv.
         args = None
+
         if os.environ.get('_ARGCOMPLETE'):
             # During argcomplete completion, parse_known_args will return an
             # empty Namespace. In this case, we feed the previous function with
@@ -257,7 +263,6 @@ class Config(object):
             args = self.parser.parse_known_args(compline.split()[1:])[0]
         else:
             args = self.parser.parse_known_args()[0]
-
         if hooks is not None:
             if isinstance(hooks, list):
                 for h in hooks:
@@ -279,10 +284,10 @@ class Config(object):
             log = logging.getLogger('argtoolbox')
             log.debug("reloading configuration ...")
             if args.config_file:
-                self.file_parser = ConfigParser.SafeConfigParser()
+                self.file_parser = configparser.SafeConfigParser()
                 discoveredFileList = self.file_parser.read(args.config_file)
                 log.debug("discoveredFileList: " + str(discoveredFileList))
-            for s in self.sections.values():
+            for s in list(self.sections.values()):
                 log.debug("loading section : " + s.get_section_name())
                 s.reset()
                 s.load(self.file_parser)
@@ -292,7 +297,7 @@ class Config(object):
         if name.lower() == "default":
             return self._default_section
         s = self.sections.get(name)
-        if s:
+        if s is not None:
             return s
         else:
             raise AttributeError("'%(class)s' object has no attribute \
@@ -301,7 +306,7 @@ class Config(object):
     def __str__(self):
         res = []
         res.append("Configuration of %(prog_name)s : " % self.__dict__)
-        for s in self.sections.values():
+        for s in list(self.sections.values()):
             res.append("".join(s.get_representation("\t")))
         return "\n".join(res)
 
@@ -324,7 +329,7 @@ class Config(object):
                         f.write("\n")
                     f.write("\n\n")
 
-                for s in self.sections.values():
+                for s in list(self.sections.values()):
                     log.debug("loading section : " + s.get_section_name())
                     s.write_config_file(f, comments)
             log.debug("config file generation completed : " + str(output))
@@ -425,15 +430,15 @@ class _Section(_AbstractSection):
         return len(self.elements)
 
     def reset(self):
-        for e in self.elements.values():
+        for e in list(self.elements.values()):
             e.reset()
 
     def load(self, file_parser):
         section = self.get_section_name()
         try:
-            for e in self.elements.values():
+            for e in list(self.elements.values()):
                 e.load(file_parser, section)
-        except ConfigParser.NoSectionError as e:
+        except configparser.NoSectionError as e:
             # pylint: disable-msg=W0621
             log = logging.getLogger('argtoolbox')
             if self._required:
@@ -444,12 +449,15 @@ class _Section(_AbstractSection):
 
     def __getattr__(self, name):
         e = self.elements.get(name)
-        if e:
+        if e is not None:
             return e
         else:
             raise AttributeError("'%(class)s' object has no attribute \
 '%(name)s'" % {"name": name, "class": self.__class__.__name__})
 
+    def get_element(self, name):
+        return self.elements.get(name)
+
     def write_config_file(self, f, comments):
         """This method write a sample file, with attributes, descriptions,
         sample values, required flags, using the configuration object
@@ -459,7 +467,7 @@ class _Section(_AbstractSection):
             return
         super(_Section, self).write_config_file(f, comments)
 
-        for e in self.elements.values():
+        for e in list(self.elements.values()):
             e.write_config_file(f, comments)
         f.write("\n")
 
@@ -482,7 +490,7 @@ class SimpleSection(_Section):
         if self.count() > 0:
             res.append(prefix + "Section "
                        + self.get_section_name().upper() + suffix)
-            for elt in self.elements.values():
+            for elt in list(self.elements.values()):
                 res.append("".join(elt.get_representation(prefix)))
         return res
 
@@ -496,7 +504,7 @@ class SubSection(_Section):
         if self.count() > 0:
             res.append(prefix + "SubSection : "
                        + self.get_section_name().upper() + suffix)
-            for elt in self.elements.values():
+            for elt in list(self.elements.values()):
                 res.append("".join(elt.get_representation(prefix)))
         return res
 
@@ -511,7 +519,7 @@ class SubSection(_Section):
         newone = type(self)()
         newone.__dict__.update(self.__dict__)
         newone.elements = OrderedDict()
-        for e in self.elements.values():
+        for e in list(self.elements.values()):
             newone.add_element(copy.deepcopy(e))
         return newone
 
@@ -529,11 +537,10 @@ class ListSection(_AbstractSection):
         section = self.get_section_name()
         try:
 
-            # TODO : ? : data = data.decode(locale.getpreferredencoding())
             for key in [item for item in file_parser.options(section)
-                        if item not in file_parser.defaults().keys()]:
+                        if item not in list(file_parser.defaults().keys())]:
                 self.elements[key] = file_parser.get(section, key)
-        except ConfigParser.NoSectionError as e:
+        except configparser.NoSectionError as e:
             # pylint: disable-msg=W0621
             log = logging.getLogger('argtoolbox')
             if self._required:
@@ -543,14 +550,14 @@ class ListSection(_AbstractSection):
                 log.debug("Missing section : " + section)
 
     def reset(self):
-        for e in self.elements.values():
+        for e in list(self.elements.values()):
             e.reset()
 
     def get_representation(self, prefix="", suffix="\n"):
         res = []
         res.append(prefix + "Section " + self._name + suffix)
 
-        for key, val in self.elements.items():
+        for key, val in list(self.elements.items()):
             a = []
             a.append(prefix)
             a.append(" - " + str(key) + " : " + str(val))
@@ -558,16 +565,6 @@ class ListSection(_AbstractSection):
             res.append("".join(a))
         return res
 
-    def __getattr__(self, name):
-
-        e = self.elements.get(name)
-        if e is not None:
-            return e
-        else:
-            raise AttributeError(
-                "'%(class)s' object has no attribute '%(name)s'"
-                % {"name": name, "class": self.__class__.__name__})
-
 
 # -----------------------------------------------------------------------------
 # warning| [R0902, Element] Too many instance attributes (13/7)
@@ -584,7 +581,7 @@ class Element(object):
     # pylint: disable-msg=R0913
     def __init__(self, name, e_type=str, required=False, default=None,
                  conf_hidden=False, conf_required=False, desc=None,
-                 hooks=None, hidden=False):
+                 hooks=None, hidden=False, e_type_exclude=False):
         """Information about how to declare a element which will be load from a
         configuration file.
 
@@ -594,6 +591,9 @@ class Element(object):
 
     - e_type -- Data type of the attribute.
 
+    - e_type_exclude -- Do not export data type to argparse. Example there is a
+    conflict between type=int and action=count for argparse parameters.
+
     - conf_required -- The current attribute must be present in the
     configuration file.
 
@@ -622,6 +622,7 @@ class Element(object):
 
         self._name = name
         self.e_type = e_type
+        self.e_type_exclude = e_type_exclude
         self._required = required
         self.default = default
         self._desc = desc
@@ -629,7 +630,7 @@ class Element(object):
         self.conf_required = conf_required
         self._desc_for_config = None
         self._desc_for_argparse = None
-        self.value = None
+        self._value = None
         self.hidden = hidden
 
         if hooks is None:
@@ -648,6 +649,18 @@ class Element(object):
                 raise TypeError(
                     "Hook argument should be a subclass of DefaultHook")
 
+    @property
+    def value(self):
+        """TODO"""
+        if self._value is None:
+            return self.default
+        return self._value
+
+    @value.setter
+    def value(self, value):
+        """TODO"""
+        self._value = value
+
     def get_name(self):
         """This method will return the name of the current element"""
         return self._name
@@ -662,12 +675,15 @@ class Element(object):
             res.append(prefix + " - " + str(self._name)
                        + " : xxxxxxxx" + suffix)
         else:
+            default = self.default
+            if default is None:
+                default = " - "
             a = prefix + " - "
             a += str(self._name) + " : "
-            if isinstance(self.value, types.UnicodeType):
-                a += self.value
+            if isinstance(default, str):
+                a += default
             else:
-                a += str(self.value)
+                a += str(default)
             a += suffix
             res.append(a)
         return res
@@ -716,7 +732,6 @@ class Element(object):
                 elif self.e_type == list:
                     data = file_parser.get(section_name, self._name)
                     data = data.strip()
-                    data = data.decode(locale.getpreferredencoding())
                     data = data.split()
                     if not data:
                         msg = "The optional field '%(name)s' was present, \
@@ -734,7 +749,6 @@ list." % {"name": self._name}
                         string." % {"name": self._name}
                         log.error(msg)
                         raise ValueError(msg)
-                    data = data.decode(locale.getpreferredencoding())
                 else:
                     msg = "Data type not supported : %(type)s " % {
                         "type": self.e_type}
@@ -759,7 +773,7 @@ required type is : %(e_type)s." %  {
 type : '%(e_type)s'", log_data)
             self.value = data
 
-        except ConfigParser.NoOptionError:
+        except configparser.NoOptionError:
             if self.conf_required:
                 msg = "The required field '%(name)s' was missing from the \
 config file." % {"name": self._name}
@@ -796,6 +810,10 @@ config file." % {"name": self._name}
             else:
                 ret["required"] = True
         ret["dest"] = self._name
+        if not self.e_type_exclude:
+            if self.e_type == int or self.e_type == float:
+                # Just override argparse.add_argument 'type' parameter for int or float.
+                ret["type"] = self.e_type
         if self.value is not None:
             ret["default"] = self.value
         if self._desc:
@@ -864,7 +882,7 @@ class ElementWithSubSections(Element):
         res.append(temp_line)
 
         if len(self.sections) > 0:
-            for elt in self.sections.values():
+            for elt in list(self.sections.values()):
                 res.append("".join(elt.get_representation(prefix + "\t")))
         return res
 
@@ -882,7 +900,7 @@ class ElementWithSubSections(Element):
     def load(self, file_parser, section_name):
         self._load(file_parser, section_name)
         if len(self.sections) > 0:
-            for sec in self.sections.values():
+            for sec in list(self.sections.values()):
                 sec.name = self.value
                 sec.load(file_parser)
         self.post_load()
@@ -966,6 +984,7 @@ class ElementWithRelativeSubSection(ElementWithSubSections):
 
     def get_representation(self, prefix="", suffix="\n"):
         res = ['\n']
+        res.append('#')
         temp_line = prefix + " - " + str(self._name) + " : "
         if self.hidden:
             temp_line += "xxxxxxxx" + suffix
@@ -973,8 +992,10 @@ class ElementWithRelativeSubSection(ElementWithSubSections):
             temp_line += str(self.value) + suffix
         res.append(temp_line)
 
+        res.append("".join(self.rss.get_representation(prefix + "#\t")))
+        res.append('\n')
         if len(self.sections) > 0:
-            for elt in self.sections.values():
+            for elt in list(self.sections.values()):
                 res.append('\n')
                 res.append("".join(elt.get_representation(prefix + "\t")))
             res.append('\n')
@@ -1002,10 +1023,10 @@ class DefaultCommand(object):
         self.log.debug("Namespace : begin :")
         for i in dict_tmp.__dict__:
             attribute = getattr(dict_tmp, i)
-            if isinstance(attribute, types.UnicodeType):
-                self.log.debug(i + " : " + attribute)
+            if isinstance(attribute, str):
+                self.log.debug(i + " : " + attribute + " : <type 'unicode'>")
             else:
-                self.log.debug(i + " : " + str(attribute))
+                self.log.debug(i + " : " + str(attribute) + " : " + str(type(attribute)))
         self.log.debug("Namespace : end.")
 
     # pylint: disable-msg=W0613
@@ -1023,22 +1044,22 @@ class TestCommand(DefaultCommand):
 
     def __call__(self, args):
         super(TestCommand, self).__call__(args)
-        print ""
-        print "This is the beginning of the TestCommand class."
-        print ""
+        print("")
+        print("This is the beginning of the TestCommand class.")
+        print("")
 
-        print "The loaded configuration : "
-        print "---------------------------"
-        print unicode(self.config)
-        print ""
+        print("The loaded configuration : ")
+        print("---------------------------")
+        print(str(self.config))
+        print("")
 
-        print "The command line arguments (argv) : "
-        print "------------------------------------"
-        print args
+        print("The command line arguments (argv) : ")
+        print("------------------------------------")
+        print(args)
 
-        print ""
-        print "This is the end of the TestCommand class."
-        print ""
+        print("")
+        print("This is the end of the TestCommand class.")
+        print("")
 
 
 # -----------------------------------------------------------------------------
@@ -1053,7 +1074,7 @@ class DefaultCompleter(object):
         try:
             debug("\n------------ DefaultCompleter -----------------")
             debug("Kwargs content :")
-            for i, j in kwargs.items():
+            for i, j in list(kwargs.items()):
                 debug("key : " + str(i))
                 debug("\t - " + str(j))
             debug("\n------------ DefaultCompleter -----------------\n")
@@ -1066,13 +1087,18 @@ class DefaultCompleter(object):
             # getting form args the current Command and looking for a method
             # called by default 'complete'. See __init__ method. The method
             # name is store in the class member called self.func_name
-            fn = getattr(args.__func__, self.func_name, None)
+            fn = getattr(args.__func__, self.__name__, None)
             if fn:
                 return fn(args, prefix)
 
         # pylint: disable-msg=W0703
         except Exception as e:
-            debug("\nERROR:An exception was caught :" + str(e) + "\n")
+            from argcomplete import warn
+            warn("\nERROR::COMPLETE:An exception was caught :" + str(e) + "\n")
+            import traceback
+            traceback.print_exc()
+            debug("\n------\n")
+            return ["comlete-error"]
 
 
 # -----------------------------------------------------------------------------
@@ -1085,8 +1111,44 @@ class DefaultProgram(object):
         self.force_debug = force_debug
 
     def __call__(self):
+
+        def patch(self, parser, namespace, values, option_string=None):
+            """patch original __call__ method for argparse 1.1 (fix)"""
+            from argparse import _UNRECOGNIZED_ARGS_ATTR
+            from argparse import SUPPRESS
+            from argparse import _
+
+            parser_name = values[0]
+            arg_strings = values[1:]
+
+            # set the parser name if requested
+            if self.dest is not SUPPRESS:
+                setattr(namespace, self.dest, parser_name)
+
+            # select the parser
+            try:
+                parser = self._name_parser_map[parser_name]
+            except KeyError:
+                tup = parser_name, ', '.join(self._name_parser_map)
+                msg = _('unknown parser %r (choices: %s)') % tup
+                raise ArgumentError(self, msg)
+
+            # parse all the remaining options into the namespace
+            # store any unrecognized options on the object, so that the top
+            # level parser can decide what to do with them
+
+            namespace, arg_strings = parser.parse_known_args(arg_strings,
+                                                             namespace)
+            if arg_strings:
+                vars(namespace).setdefault(_UNRECOGNIZED_ARGS_ATTR, [])
+                getattr(namespace, _UNRECOGNIZED_ARGS_ATTR).extend(arg_strings)
+
         # integration with argcomplete python module (bash completion)
         try:
+            # Very ugly dirty hook/fix
+            if argparse.__version__ == "1.1":
+                # pylint: disable=protected-access
+                argparse._SubParsersAction.__call__ = patch
             import argcomplete
             argcomplete.autocomplete(self.parser)
         except ImportError:
@@ -1096,28 +1158,37 @@ class DefaultProgram(object):
         args = self.parse_args()
 
         if getattr(args, 'debug', False) or self.force_debug:
-            llog = logging.getLogger()
-            llog.setLevel(logging.DEBUG)
+            if hasattr(args, 'debugger_names') and getattr(args, 'debugger_names'):
+                for name in args.debugger_names:
+                    llog = logging.getLogger(name)
+                    llog.setLevel(logging.DEBUG)
+            else:
+                llog = logging.getLogger()
+                llog.setLevel(logging.DEBUG)
             streamHandler.setFormatter(DEBUG_LOGGING_FORMAT)
             if self.config:
-                print "debug>>>----------- config -------------------"
-                print unicode(self.config)
-                print "debug----------- processing --------------<<<<"
+                print("debug>>>----------- config -------------------")
+                print(str(self.config))
+                print("debug----------- processing --------------<<<<")
 
             # run command
-            args.__func__(args)
-            return True
+            if not hasattr(args, '__func__'):
+                self.parser.error("You must provide a command. See --help.")
+            return args.__func__(args)
         else:
             # pylint: disable-msg=W0621
             log = logging.getLogger('argtoolbox')
             try:
                 # run command
-                args.__func__(args)
-                return True
+                if not hasattr(args, '__func__'):
+                    self.parser.error("You must provide a command. See --help.")
+                return args.__func__(args)
             except ValueError as a:
                 log.error("ValueError : " + str(a))
             except KeyboardInterrupt as a:
                 log.warn("Keyboard interruption detected.")
+            except ArgumentError as a:
+                self.parser.error(a.message)
             except Exception as a:
                 log.error("unexcepted error : " + str(a))
             return False
@@ -1128,14 +1199,13 @@ class DefaultProgram(object):
         for attr in args.__dict__:
             data = getattr(args, attr)
             if isinstance(data, str):
-                data = data.decode(locale.getpreferredencoding())
                 setattr(args, attr, data)
 
             if isinstance(data, list):
                 res = []
                 for val in data:
                     if isinstance(val, str):
-                        res.append(val.decode(locale.getpreferredencoding()))
+                        res.append(val)
                     else:
                         res.append(val)
                 setattr(args, attr, res)
@@ -1194,6 +1264,7 @@ class BasicProgram(object):
         default.add_element(
             Element('debug',
                     e_type=int,
+                    e_type_exclude=True,
                     default=0,
                     desc="""debug level : default : 0."""))
 
@@ -1296,16 +1367,16 @@ def _query_yes_no(question, default="yes"):
     else:
         raise ValueError("invalid default answer: '%s'" % default)
 
-    while 1:
+    while True:
         sys.stdout.write(question + prompt)
         try:
-            choice = raw_input().lower()
+            choice = input().lower()
         except KeyboardInterrupt:
-            print
+            print()
             return "no"
         if default is not None and choice == '':
             return default
-        elif choice in valid.keys():
+        elif choice in list(valid.keys()):
             return valid[choice]
         else:
             sys.stdout.write("Please respond with 'yes' or 'no' "
